#    Copyright 2004 Andrew Wilkinson <aw@cs.york.ac.uk>.
#
#    This file is part of PyLinda (http://www-users.cs.york.ac.uk/~aw/pylinda)
#
#    PyLinda is free software; you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation; either version 2.1 of the License, or
#    (at your option) any later version.
#
#    PyLinda is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with PyLinda; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

## \namespace utils
## \brief This module contains all the various functions that are shared amongst the linda modules
##
## \internal

import thread
import struct

import typeconvert
encode = typeconvert.pack
decode = typeconvert.unpack

max_msg_size = 1024
def recv(s):
    """\internal
    \brief Receive a message from the given socket

    \param s A socket object
    """
    msg = ""
    while len(msg) < 2:
        msg += s.recv(max_msg_size)
        if msg == "":
            return ""

    size, msg = struct.unpack("!H", msg[:2])[0], msg[2:]

    if size > len(msg):
        while size > len(msg):
            msg += s.recv(size - len(msg))

    return decode(msg)

def send(s, msg):
    """\internal
    \brief Sends a message on the given socket

    \param s A socket object
    """
    #s.send(encode(msg))
    msg = encode(msg)
    s.send(struct.pack("!H", len(msg)) + msg)

def containsTS(tup, func):
    """\internal
    \brief Recursively scans the given tuple (tup) for a tuple space and calls func with each tuple space that is found as an argument

    This function is very useful when used in conjuction with map.
    \code
    map(list_of_tuples, lambda x: utils.containsTS(x, func))
    \endcode

    \param tup The tuple to be checked
    \param func The function to be called for each Tuplespace reference that is found
    """
    map(lambda t: containsTS(t, func), filter(lambda t: t.__class__ == tuple, tup))
    map(func, filter(lambda t: isinstance(t, TupleSpace), tup))

def changeOwner(t, to):
    """\internal
    \brief Changes the owner of the given tuplespace reference to the given process/server.

    \param t A tuplespace reference
    \param to A process/server id
    """
    t._addreference(to)
    t._delreference(t.owner)
    t.owner = to

def isTupleSpaceId(id):
    """\internal
    \brief Checks if the given id is a valid tuplespace id

    This function does not check that a tuplespace exists with this id, only the id is of the correct format
    \param id The id to be checked
    """
    try:
        if type(id) != str:
            return False
        node, num = id.split(":")
        int(node)
        int(num)
    except ValueError:
        return False
    else:
        return True

def isNodeId(id):
    """\internal
    \brief Checks if the given id is a valid node id

    This function does not check that a node exists with this id, only the id is of the correct format
    \param id The id to be checked
    """
    return type(id) == int

def isProcessId(id):
    """\internal
    \brief Checks if the given id is a valid process id

    This function does not check that a process exists with this id, only the id is of the correct format
    \param id The id to be checked
    """
    try:
        if type(id) != str:
            return False
        pid, num = id.split("!")
        int(pid)
        int(num)
    except ValueError:
        return False
    else:
        return True

def getNodeFromTupleSpaceId(ts):
    """\internal
    \brief Returns the node that owns the given tuplespace

    \param ts The tuplespace whoes owner you want
    \return The node who owns the tuplespace
    """
    assert isTupleSpaceId(ts), "Not a tuplespace id %s" % (ts, )

    if int(ts.split(":")[0]) == 0:
        return 1
    else:
        return int(ts.split(":")[0])

def getNodeFromProcessId(ts):
    """\internal
    \brief Returns the node that owns the given process

    \param ts The process whos owner you want
    \return The node who owns the tuplespace
    """
    assert isProcessId(ts), "Not a process id %s" % (ts, )

    if int(ts.split("!")[0]) == 0:
        return 1
    else:
        return int(ts.split("!")[0])

class Counter:
    def __init__(self, start=0):
        self.counter = start
        self.lock = thread.allocate_lock()

    def __call__(self):
        self.lock.acquire()
        self.counter = self.counter + 1
        counter = self.counter
        self.lock.release()
        return counter

    def next(self):
        return self()
